"""Module for analysis of static network data."""

from warnings import warn

import networkx as nx

from entfrachten.analysis.drainage_sysytem_data import DrainageSystemData

def in_out_reach_subset(sd:DrainageSystemData,in_reaches:list[str],out_reaches:list[str]) -> tuple[list[str],list[str]]:
    """Find all nodes and reaches between in_reaches and out_reaches."""
    # we want to look at everything between the in/out reaches.
    # the nodes where inflow starts are not included, so we take end here instead of start.
    in_nodes = sd.reach_df.loc[in_reaches, "end"]
    out_nodes = sd.reach_df.loc[out_reaches, "start"]
    # make a graph where in and out reaches are removed, the nodes connected to in or out are then the subgraph we need.
    cut_reaches = sd.reach_df.copy()
    cut_reaches = cut_reaches.loc[sd.reach_df.index.difference(in_reaches).difference(out_reaches)]
    # also remove weirs, since we should not count parts where the water overflows.
    cut_reaches = cut_reaches.query("type != 'Weir'")
    cut_graph = nx.from_pandas_edgelist(
        cut_reaches.reset_index(),
        source="start",
        target="end",
        edge_attr=True,
        create_using=nx.DiGraph,
    )
    # Add in/out nodes to make sure that the next calculations run without error.
    # They may be missing if in/outreaches and weirs are the only reaches that connect there.
    cut_graph.add_nodes_from(in_nodes)
    cut_graph.add_nodes_from(out_nodes)
    
    in_downstream = set()
    for ni in in_nodes:
        in_downstream.update(nx.descendants(cut_graph, ni))
        in_downstream.add(ni)
    out_upstream = set()
    for no in out_nodes:
        out_upstream.update(nx.ancestors(cut_graph, no))
        out_upstream.add(no)
    for r, ni in in_nodes.items():
        if not ni in out_upstream:
            warn(f"in_node {ni} (in_reach {r}) not connected to out")
    for r, no in out_nodes.items():
        if not no in in_downstream:
            warn(f"out_node {no} (out_reach {r}) not connected to in")
    nodes = list(in_downstream | out_upstream)
    subgraph = cut_graph.subgraph(nodes)
    reaches = set()
    reaches.update(in_reaches)
    reaches.update(out_reaches)
    reaches.update(nx.get_edge_attributes(subgraph, "name").values())
    reaches = list(reaches)
    return nodes, reaches