"""Plots for structure analysis."""

import typing
import numpy as np
import pathlib
import pandas as pd
import scipy.stats
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm import tqdm

from entfrachten.analysis import (
    drainage_system_graph,
    drainage_sysytem_data,
    unit_factors as uf,
    dynamic_analysis as da,
    sewer_map,
)

# Helper for more concise indexing
ids = pd.IndexSlice

fig_height = 8 * uf.CM_TO_INCH
font_size = 9

# common arguments for boxplots
box_kwargs = dict(
    whis=(10.0, 90.0),
    showmeans=True,
    meanline=True,
    meanprops=dict(linestyle=":", linewidth=1.2, color="black"),
)


def adjust_mpl_settings():
    """Adjusts matplotlib settings. Should be called before plotting."""
    mpl.rcParams.update(
        {
            "font.size": font_size,  # Set font size
            "font.family": "Arial",  #  and font family
            "figure.figsize": [fig_height, fig_height],
        }
    )


def dist_line(
    df_dist: pd.DataFrame, out_folder: pathlib.Path, quantity: str, unit: str
):
    """Draws line plots for all places and events. Uses 'part' level for color. Intended for analysis of inflow,outflow,overflow and storage"""
    quantity_label = f"{quantity} ({unit})"
    max_events_per_plot = 16
    line_folder = out_folder / "lines/"
    line_folder.mkdir(exist_ok=True)
    for p in tqdm(df_dist.index.get_level_values("place").unique()):
        df_dist_p = df_dist.xs(p, level="place")
        events_p = sorted(df_dist_p.index.get_level_values("event").unique())
        event_chunks = [
            events_p[i : i + max_events_per_plot]
            for i in range(0, len(events_p), max_events_per_plot)
        ]
        for ec in tqdm(event_chunks):
            e_min = min(ec)
            e_max = max(ec)
            title = f"{quantity} at {p} events {e_min}-{e_max}"
            filename = f"{p.replace(' ','')}_lines_{quantity.replace(' ','_')}_{e_min}-{e_max}.svg"
            plot_df = (
                df_dist_p.query("event in @ec")
                .stack()
                .rename(quantity_label)
                .reset_index()
            )
            fg = (
                sns.FacetGrid(
                    plot_df.reset_index(),
                    col="event",
                    col_wrap=4,
                    hue="part",
                    palette=_dist_palette(df_dist_p),
                    sharex=False,
                    sharey=False,
                    height=fig_height,
                    aspect=2.7,
                )
                .map(sns.lineplot, "datetime", quantity_label, alpha=0.6)
                .add_legend()
            )
            fg.figure.suptitle(title, y=1.04)
            path = line_folder / filename
            print(f"saving to {path}")
            fg.savefig(path)


def dist_box_max(
    df_dist: pd.DataFrame, out_folder: pathlib.Path, quantity: str, unit: str
):
    """One Plot per place. Draws box plots for of the max values per event. Uses 'part' level for color.
    Intended for analysis of inflow,outflow,overflow and storage"""
    # boxplots per place of max values
    for p, df_dist_p in df_dist.groupby("place"):
        assert isinstance(p, str)
        # drop columns with only nan, this should be the distribute_storage columns for places where they do not apply
        df_dist_p = df_dist_p.dropna(how="all", axis="columns")
        title = f"max {quantity} parts at {p}, absolute"
        filename = f"{p.replace(' ','')}_{quantity.replace(' ','_')}_max_box.svg"
        plot_df = (
            df_dist_p.groupby("event")
            .max()
            .stack()
            .rename(f"{quantity} ({unit})")
            .reset_index()
        )
        fg = (
            sns.FacetGrid(
                plot_df.reset_index(),
                hue="part",
                palette=_dist_palette(df_dist),
                sharex=False,
                sharey=False,
                height=fig_height,
                aspect=1.5,
            )
            .map(
                sns.boxplot,
                "part",
                f"{quantity} ({unit})",
                **box_kwargs,
                order=list(df_dist_p.columns),
            )
            .add_legend()
        )
        plt.suptitle(title, y=1.04)
        plt.show()
        path = out_folder / filename
        # print(f"saving to {path}")
        fg.savefig(path)


def dist_in_volume_section(
    df_dist_d: pd.DataFrame,
    sealed_area_m2: pd.Series,
    quantity: str,
    unit: str,
    out_folder: pathlib.Path,
):
    """One plot per place. Draws box plots and line plots of the median for of the values per volume section. Uses 'part' level for color.
    Intended for analysis of inflow,outflow,overflow and storage."""

    box_kw = box_kwargs | dict(fliersize=2, native_scale=True)
    _dist_in_volume_section_impl(
        f"{quantity}_dist_box",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        True,
        sns.boxplot,
        box_kw,
    )
    _dist_in_volume_section_impl(
        f"{quantity}_dist_box",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        False,
        sns.boxplot,
        box_kw,
    )
    line_kw = dict(
        estimator="median",
        errorbar=("pi", 50),
        marker=".",
        markersize=6,
        err_style="band",
    )
    _dist_in_volume_section_impl(
        f"{quantity}_dist_quartline",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        True,
        sns.lineplot,
        line_kw,
    )
    _dist_in_volume_section_impl(
        f"{quantity}_dist_quartline",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        False,
        sns.lineplot,
        line_kw,
    )


def _dist_in_volume_section_impl(
    plot_name,
    df_dist_d: pd.DataFrame,
    sealed_area_m2: pd.Series,
    quantity: str,
    unit: str,
    out_folder: pathlib.Path,
    filter_overflow: bool,
    plot_fkt: typing.Callable,
    fkt_kwargs: dict,
):
    """implementation for dist_in_volume_section. The plot function is just an argument,
    most of the code is for preparing the data and is the same for line or box plots."""
    for p in tqdm(df_dist_d.index.get_level_values("place").unique()):
        df_dist_p = df_dist_d.xs(p, level="place")
        df_dist_p = df_dist_p.dropna(how="all", axis="columns")
        section_size = (
            df_dist_p.index.get_level_values("volume_section")
            .to_series()
            .diff()
            .median()
        )
        title = f"{p}: {quantity} change over incoming volume sections({section_size:.2f} m³)"
        filename = f"{p.replace(' ','')}_{plot_name}"
        if filter_overflow:
            # filter out events where no overflow occurs
            events_p = sorted(df_dist_p.index.get_level_values("event").unique())
            events_p = [
                e for e in events_p if df_dist_p.loc[ids[e, :], "overflow"].max() > 0
            ]
            if len(events_p) == 0:
                print("no events with overflow for " + p)
                continue
            df_dist_p = df_dist_p.query("event in @events_p")
            title += ", only overflow events"
            filename += "_overflows_only"
        filename += ".svg"
        # stack dataframe such that a volume_in exists for every row while other parts are stackes on top of each
        parts = list(df_dist_p.columns)
        df_dist_p = df_dist_p.reset_index()
        rain_depth = da.calc_raindepth(df_dist_p["volume_section"], sealed_area_m2[p])
        plot_df = pd.concat(
            [
                pd.DataFrame(
                    {
                        "volume section(mm)": rain_depth,
                        f"Δ {quantity} ({unit})": df_dist_p[part],
                        "part": part,
                    }
                )
                for part in parts
            ]
        )
        fg = sns.FacetGrid(
            data=plot_df,
            col="part",
            col_order=parts,
            hue="part",
            palette=_dist_palette(df_dist_d),
            hue_order=parts,
            sharey=False,
            sharex=False,
            height=fig_height,
            aspect=1,
        )
        fg.map_dataframe(
            plot_fkt, x="volume section(mm)", y=f"Δ {quantity} ({unit})", **fkt_kwargs
        )
        # adjust ylim for substorages to be equal
        sub_parts = df_dist_p.columns.difference(da.dist_agg_parts)
        sub_min = df_dist_p["storage"].min(axis=None)
        sub_max = df_dist_p["storage"].max(axis=None)
        pad = 0.05
        sub_ylim = (sub_min, sub_max) # previously had *subrange for both limits
        for part, ax in fg.axes_dict.items():
            if part in sub_parts:
                ax.set_ylim(sub_ylim)
        fg.add_legend()
        plt.suptitle(title, y=1.04) 
        path = out_folder / filename
        # print(f"saving to {path}")
        fg.savefig(path)

def dist_in_volume_section_SST(
    df_dist_d: pd.DataFrame,
    sealed_area_m2: pd.Series,
    quantity: str,
    unit: str,
    out_folder: pathlib.Path,
):
    """One plot per place. Draws box plots and line plots of the median for of the values per volume section. Uses 'part' level for color.
    Intended for analysis of inflow,outflow,overflow and storage."""

    box_kw = box_kwargs | dict(fliersize=2, native_scale=True)
    _dist_in_volume_section_impl_SST(
        f"{quantity}_dist_box",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        True,
        sns.boxplot,
        box_kw,
    )
    _dist_in_volume_section_impl_SST(
        f"{quantity}_dist_box",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        False,
        sns.boxplot,
        box_kw,
    )
    line_kw = dict(
        estimator="median",
        errorbar=("pi", 50),
        marker=".",
        markersize=6,
        err_style="band",
    )
    _dist_in_volume_section_impl_SST(
        f"{quantity}_dist_quartline",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        True,
        sns.lineplot,
        line_kw,
    )
    _dist_in_volume_section_impl_SST(
        f"{quantity}_dist_quartline",
        df_dist_d,
        sealed_area_m2,
        quantity,
        unit,
        out_folder,
        False,
        sns.lineplot,
        line_kw,
    )



def _dist_in_volume_section_impl_SST(
    plot_name,
    df_dist_d: pd.DataFrame,
    sealed_area_m2: pd.Series,
    quantity: str,
    unit: str,
    out_folder: pathlib.Path,
    filter_overflow: bool,
    plot_fkt: typing.Callable,
    fkt_kwargs: dict,
):
    """implementation for dist_in_volume_section. The plot function is just an argument,
    most of the code is for preparing the data and is the same for line or box plots."""
    for p in tqdm(df_dist_d.index.get_level_values("place").unique()):
        df_dist_p = df_dist_d.xs(p, level="place")
        df_dist_p = df_dist_p.dropna(how="all", axis="columns")
        section_size = (
            df_dist_p.index.get_level_values("volume_section")
            .to_series()
            .diff()
            .median()
        )
        title = f"{p}: {quantity} change over incoming volume sections({section_size:.2f} m³)"
        filename = f"{p.replace(' ','')}_{plot_name}"
        if filter_overflow:
            # filter out events where no overflow occurs
            events_p = sorted(df_dist_p.index.get_level_values("event").unique())
            events_p = [
                e for e in events_p if df_dist_p.loc[ids[e, :], "overflow"].max() > 0
            ]
            if len(events_p) == 0:
                print("no events with overflow for " + p)
                continue
            df_dist_p = df_dist_p.query("event in @events_p")
            title += ", only overflow events"
            filename += "_overflows_only"
        filename += "publication.svg"
        
        # Drop the 'overflow' column here
        df_dist_p = df_dist_p.drop(columns=["overflow"], errors='ignore')

        # stack dataframe such that a volume_in exists for every row while other parts are stacked on top of each
        parts = list(df_dist_p.columns)
        df_dist_p = df_dist_p.reset_index()
        rain_depth = da.calc_raindepth(df_dist_p["volume_section"], sealed_area_m2[p])
        plot_df = pd.concat(
            [
                pd.DataFrame(
                    {
                        "volume section(mm)": rain_depth,
                        f"Δ {quantity} ({unit})": df_dist_p[part],
                        "part": part,
                    }
                )
                for part in parts
            ]
        )
        fg = sns.FacetGrid(
            data=plot_df,
            col="part",
            col_order=parts,
            hue="part",
            palette=_dist_palette(df_dist_d),
            hue_order=parts,
            sharey=False,
            sharex=False,
            height=fig_height,
            aspect=1,
            col_wrap=3,  # Adjust the number of columns (3) to arrange the plots in 2 rows
        )
        fg.map_dataframe(
            plot_fkt, x="volume section(mm)", y=f"Δ {quantity} ({unit})", **fkt_kwargs
        )
        # adjust ylim for substorages to be equal
        sub_parts = df_dist_p.columns.difference(da.dist_agg_parts)
        sub_min = df_dist_p["storage"].min(axis=None)
        sub_max = df_dist_p["storage"].max(axis=None)
        sub_ylim = (sub_min, sub_max)  # previously had *subrange for both limits
        for part, ax in fg.axes_dict.items():
            if part in sub_parts:
                ax.set_ylim(sub_ylim)
        fg.set_xlabels("volume section (mm)")
        #fg.add_legend()
        # plt.suptitle(title, y=1.04) 
        path = out_folder / filename
        # print(f"saving to {path}")
        fg.savefig(path)

        
        
def _dist_palette(df) -> dict[str, str]:
    """color palette for plotting the parts"""
    part_colors = {
        "sum": "#AAAA00",
        "inflow": "#6A6A6A",
        "storage": "#328529",
        "overflow": "#BF375A",
        "outflow": "#3F63D0",
    }
    dist_distributed_cols = df.columns.difference(part_colors.keys())
    for c in dist_distributed_cols:
        part_colors[c] = part_colors["storage"]
    return part_colors


def diff_plot1(
    up_storages: dict[str, set[str]],
    overflow_filtered: pd.DataFrame,
    storage_afs: pd.DataFrame,
    timestep: pd.Timedelta,
    diff_afs_thresh: tuple[float,float], # changed to tuple for perc and abs deviation
    out_folder: pathlib.Path,
):
    """First version of plot showing differences in concentration from up to downstream."""
    # note that overflow_filtered contains ony the times where an overflow occurs, getting the timestep from it may not work.
    for down, all_ups in up_storages.items():
        for up in all_ups:
            if not (down in overflow_filtered.index.get_level_values("place")):
                print("no overflow data", down)
                continue
            overflow_down = overflow_filtered.xs(down, level="place")
            overflow_idx_down = overflow_down.index
            if len(overflow_idx_down) == 0:
                print("no overflow for", down)
            storage_up_afs = storage_afs.xs(up, level="place").loc[
                overflow_idx_down, "afs"
            ]
            storage_down_afs = storage_afs.xs(down, level="place").loc[
                overflow_idx_down, "afs"
            ]
            storage_diff_afs = storage_down_afs - storage_up_afs
            storage_diff_afs_rel = 1-(storage_up_afs/storage_down_afs)
            diff_label = f"{down}\n-{up}"
            # load that would not be in a overflow if, in times where the concentration upstream is lower,
            # the overflow would happen upstream instead of downstream.
            is_up_lower = storage_up_afs < storage_down_afs
            load_current = (
                overflow_down["afs_transport"].sum() * timestep.total_seconds()
            )
            load_diff = (
                (
                    storage_diff_afs[is_up_lower]
                    * overflow_down.loc[is_up_lower, "discharge"]
                ).sum()
                * uf.MGPL_TO_KGPM3
                * timestep.total_seconds()
            )
            overflow_duration = len(overflow_idx_down) * timestep
            # calculate afs load diff with threshold value
            up_significant_lower = (storage_diff_afs >= diff_afs_thresh[0]) & (storage_diff_afs_rel >= diff_afs_thresh[1])
            load_diff_signifcant = (
                (
                    storage_diff_afs[up_significant_lower]
                    * overflow_down.loc[up_significant_lower, "discharge"]
                ).sum()
                * uf.MGPL_TO_KGPM3
                * timestep.total_seconds()
            )
            overflow_duration = len(overflow_idx_down) * timestep            
            # time duration where the concentration difference is greater than some threshold.
            diff_greater_duration = (up_significant_lower).sum() * timestep
                        
            duration_text = f"""total overflow duration at {down}: {overflow_duration}
            tss concentration difference >= {diff_afs_thresh[0]} & >= {diff_afs_thresh[1]*100}% : {diff_greater_duration}({round(100*diff_greater_duration/overflow_duration)}%)
            Potential overflow load difference: {load_diff:.2f}kg ({round(100*load_diff/load_current)}% of {load_current:.2f}kg)
            tss concentration difference >= {diff_afs_thresh[0]} & >= {diff_afs_thresh[1]*100}% : {load_diff_signifcant:.2f}kg ({round(100*load_diff_signifcant/load_current)}% of {load_current:.2f}kg))"""

            print("making plot for ", up, "->", down)
            plot_df = pd.concat(
                [
                    pd.DataFrame(
                        {
                            "TSS (mg/l)": storage_up_afs,
                            "part": "inflow",
                            "place": up,
                        }
                    ),
                    pd.DataFrame(
                        {
                            "TSS (mg/l)": storage_diff_afs,
                            "part": "inflow",
                            "place": diff_label,
                        }
                    ),
                    pd.DataFrame(
                        {
                            "TSS (mg/l)": storage_down_afs,
                            "part": "inflow",
                            "place": down,
                        }
                    ),
                ]
            ).reset_index()
            fig, ax = plt.subplots(figsize=(fig_height, fig_height))
            ax.axhline(0, color="gray")
            sns.boxplot(
                plot_df,
                x="place",
                y="TSS (mg/l)",
                hue="place",
                fliersize=2,
                order=[up, diff_label, down],
                **box_kwargs,
                ax=ax,
            )
            ax.set(xlabel=None)
            fig.text(
                -0.1,
                -0.07,
                duration_text,
                horizontalalignment="left",
                verticalalignment="top",
                fontsize="medium",
            )
            fig.suptitle(f"TSS concentration during overflow at {down}", y=1.04)
            fig.savefig(out_folder / f"diff_conc_{up}-{down}.svg", bbox_inches="tight")
            fig, ax = plt.subplots(figsize=(fig_height, fig_height))
            ax.axhline(0, color="gray")
            sns.violinplot(
                plot_df,
                x="place",
                hue="place",
                y="TSS (mg/l)",
                inner=None,
                order=[up, diff_label, down],
                ax=ax,
            )
            ax.set(xlabel=None)
            plt.suptitle(f"TSS concentration during overflow at {down}", y=1.04)
            fig.text(
                -0.1,
                -0.07,
                duration_text,
                horizontalalignment="left",
                verticalalignment="top",
                fontsize="medium",
            )
            fig.savefig(
                out_folder / f"diff_conc_violin{up}-{down}.svg", bbox_inches="tight"
            )


def diff_plot2(
    up_storages: dict[str, set[str]],
    overflow_filtered: pd.DataFrame,
    inflow: pd.DataFrame,
    outflow: pd.DataFrame,
    storage_afs: pd.DataFrame,
    storage_struct: pd.DataFrame,
    graph: drainage_system_graph.DrainageSystemGraph,
    uds: drainage_sysytem_data.DrainageSystemData,
    timestep: pd.Timedelta,
    diff_afs_thresh: tuple[float,float], #changed to tuple
    out_folder: pathlib.Path,
):
    """Second version of plot showing differences in concentration from up to downstream.
    Accounts for difference in flow times and shows potential for overflow load reduction.
    Also creates line plots for some events to show the time shift."""
    for down, all_ups in up_storages.items():
        for up in all_ups:
            up_node = storage_struct.loc[up, "node"]
            down_node = storage_struct.loc[down, "node"]
            path = graph.find_shortest_path(up_node, down_node)
            sewer_map.make_path_time_map(uds, path).save(
                out_folder / f"diff_conc2_{up}-{down}_flowtime.html"
            )
            flow_len = uds.reach_df.loc[path, "length"].sum()
            flow_time = uds.reach_df.loc[path, "flow_time"].sum()
            example_lines_shifted(down, up, flow_time, inflow, out_folder)
            if not (down in overflow_filtered.index.get_level_values("place")):
                print("no overflow data", down)
                continue
            overflow_down = overflow_filtered.xs(down, level="place")
            overflow_idx_down = overflow_down.index
            if len(overflow_idx_down) == 0:
                print("no overflow for", down)
            # we need to account for flow time.
            # water that could overflow at up would only get to down after flow_time has passed
            # therefore we need to get the values from up at t-flow_time.
            # to operate on the index values, first convert to dataframe
            overflow_idx_up = overflow_idx_down.to_frame()
            # we need to round flow_time to match the timestep of our data
            overflow_idx_up["datetime"] -= flow_time.round(timestep)
            overflow_idx_up = pd.MultiIndex.from_frame(overflow_idx_up)
            storage_up_afs = storage_afs.xs(up, level="place").reindex(overflow_idx_up)[
                "afs"
            ]
            # overwrite index, otherwise pandas would match the original times.
            storage_up_afs.index = overflow_idx_down
            storage_down_afs = storage_afs.xs(down, level="place").loc[
                overflow_idx_down, "afs"
            ]
            storage_diff_afs = storage_down_afs - storage_up_afs
            storage_diff_afs_rel = 1-(storage_up_afs/storage_down_afs)
            diff_label = f"{down}\n-{up}"
            # load that would not be in a overflow if, in times where the concentration upstream is lower,
            # the overflow would happen upstream instead of downstream.
            is_up_lower = storage_up_afs < storage_down_afs
            load_current = (
                overflow_down["afs_transport"].sum() * timestep.total_seconds()
            )
            # the overflow that is shifted to up cannot be greater than the flow that comes from there
            # TODO: flow_up ändern zum maximalen aufgezeichneten Abfluss
            flow_up = outflow.xs(up, level="place").reindex(overflow_idx_up)[
                "discharge"
            ]
            flow_up.index = overflow_idx_down
            flow_shiftable = overflow_down["discharge"].clip(upper=flow_up)
            
            load_diff = (
                (storage_diff_afs[is_up_lower] * flow_shiftable.loc[is_up_lower]).sum()
                * uf.MGPL_TO_KGPM3
                * timestep.total_seconds()
            )
            overflow_duration = len(overflow_idx_down) * timestep
            # calculate afs load diff with threshold value
            up_significant_lower = (storage_diff_afs >= diff_afs_thresh[0]) & (storage_diff_afs_rel >= diff_afs_thresh[1])
            load_diff_signifcant = (
                (
                    storage_diff_afs[up_significant_lower]
                    * flow_shiftable.loc[up_significant_lower]
                ).sum()
                * uf.MGPL_TO_KGPM3
                * timestep.total_seconds()
            )
            overflow_duration = len(overflow_idx_down) * timestep            
            # time duration where the concentration difference is greater than some threshold.
            diff_greater_duration = up_significant_lower.sum() * timestep
            note_text = f"""total overflow duration at {down}: {overflow_duration}
    tss concentration difference >= {diff_afs_thresh[0]} & >= {diff_afs_thresh[1]*100}% : {diff_greater_duration}({round(100*diff_greater_duration/overflow_duration)}%)
    Potential overflow load difference: {load_diff:.2f}kg ({round(100*load_diff/load_current)}% of {load_current:.2f}kg)
    tss concentration difference >= {diff_afs_thresh[0]} & >= {diff_afs_thresh[1]*100}% : {load_diff_signifcant:.2f}kg ({round(100*load_diff_signifcant/load_current)}% of {load_current:.2f}kg)
    flow time: {flow_time}s, flow length: {flow_len:.2f}m, missing upstream data:{round(100*storage_up_afs.isna().mean())}%"""
            print("making plot for ", up, "->", down)
            plot_df = pd.concat(
                [
                    pd.DataFrame(
                        {
                            "TSS (mg/l)": storage_up_afs,
                            "place": up,
                        }
                    ),
                    pd.DataFrame(
                        {
                            "TSS (mg/l)": storage_diff_afs,
                            "place": diff_label,
                        }
                    ),
                    pd.DataFrame(
                        {
                            "TSS (mg/l)": storage_down_afs,
                            "place": down,
                        }
                    ),
                ]
            ).reset_index()
            fig, (ax, ax_zoom) = plt.subplots(
                ncols=2,
                figsize=(fig_height * 2.5, fig_height),
                gridspec_kw=dict(wspace=0.24),
            )
            ax.axhline(0, color="gray")
            sns.boxplot(
                plot_df,
                x="place",
                y="TSS (mg/l)",
                hue="place",
                fliersize=2,
                order=[up, diff_label, down],
                **box_kwargs,
                ax=ax,
            )
            ax.set(xlabel=None)
            ax.axhline(0, color="gray")
            sns.boxplot(
                plot_df,
                x="place",
                y="TSS (mg/l)",
                hue="place",
                showfliers=False,
                order=[up, diff_label, down],
                **box_kwargs,
                ax=ax_zoom,
            )
            ax_zoom.set(xlabel=None)
            ax_zoom.axhline(0, color="gray")
            fig.text(
                0.2,
                -0.07,
                note_text,
                horizontalalignment="left",
                verticalalignment="top",
                fontsize="medium",
            )
            fig.suptitle(f"TSS concentration during overflow at {down}", y=1.04)
            fig.savefig(out_folder / f"diff_conc2_{up}-{down}.svg", bbox_inches="tight")
            plt.close(fig) 
            
            #create a new graph for publication
            anonymize_dict = {
                    "SKU 0901": "CSO1", "SKU 0216": "CSO2", "SKU 0222": "CSO3",
                    "SKU 0225": "CSO4", "SKU 0228": "CSO5", "RUE 0139": "CSO6",
                    "MWE 0145": "CSO7", "SKU 0148": "CSO8", "SKU 0147": "CSO9",
                    "SKU 0133": "CSO10", "RKB 0337": "SSO1", "RRB 0301": "SST1"
                }
            # Rename up and down before using them
            up_anon = anonymize_dict.get(up, up)     # Default to original value if not found
            down_anon = anonymize_dict.get(down, down)
            
            # Create the label using renamed values
            diff_label2 = f"{down_anon}\n-{up_anon}"
            
            # Create the DataFrame with renamed values
            plot_df2 = pd.concat(
                [
                    pd.DataFrame({"TSS (mg/l)": storage_up_afs, "place": up_anon}),
                    pd.DataFrame({"TSS (mg/l)": storage_diff_afs, "place": diff_label2}),
                    pd.DataFrame({"TSS (mg/l)": storage_down_afs, "place": down_anon}),
                ]
            ).reset_index(drop=True)
            
            fig, ax = plt.subplots(
                ncols=1,
                figsize=(fig_height, fig_height),
                gridspec_kw=dict(wspace=0.24),
            )
            ax.set(xlabel=None)
            ax.axhline(0, color="gray")
            sns.boxplot(
                plot_df2,
                x="place",
                y="TSS (mg/l)",
                #hue="place",
                showfliers=False,
                order=[up_anon, diff_label2, down_anon],
                **box_kwargs,
                ax=ax,
            )
            fig.savefig(out_folder / f"diff_conc2_{up}-{down}_publication.svg", bbox_inches="tight")            
            plt.close(fig) 
            
            # Create a dictionary for storing data of the printed text to an excel file
            data_dict = {
                "up":up,
                "down": down,
                "flow_time (h)": flow_time.total_seconds()/60/60,
                "flow_length_m": flow_len,
                "overflow_duration (h)": overflow_duration.total_seconds()/60/60,
                "threshold absolute": diff_afs_thresh[0],
                "threshold percentage": diff_afs_thresh[1] * 100,
                "diff >0 (h)": diff_greater_duration.total_seconds()/60/60,
                "diff >0 (%)": round(100 * diff_greater_duration / overflow_duration),
                "overflow load": load_current,
                "relocatable load (kg)": load_diff,
                "load_diff_percent": round(100 * load_diff / load_current),
                "relocatable load significant (kg)": load_diff_signifcant,
                "load_diff_significant_percent": round(100 * load_diff_signifcant / load_current),
                "missing_upstream_data_percent": round(100 * storage_up_afs.isna().mean()),
            }
            
            # Convert to a DataFrame
            df_summary = pd.DataFrame([data_dict])
            
            # Save as CSV (or Excel if preferred)
            out_path = out_folder / f"diff_conc2_{up}-{down}.xlsx"
            df_summary.to_excel(out_path, index=False)


def example_lines_shifted(down, up, up_shift, inflow, out_folder):
    """Creates line plots for some events to show the time shift."""
    example_events = [363, 505, 588, 320, 490, 563]
    discharge_down = inflow.query("place==@down and event in @example_events")[
        "discharge"
    ].reset_index()
    discharge_up = inflow.query("place==@up and event in @example_events")[
        "discharge"
    ].reset_index()
    discharge_up_shifted = discharge_up.copy()
    discharge_up_shifted["datetime"] = discharge_up_shifted["datetime"] + up_shift
    discharge_up["place"] += "_original"
    discharge_up_shifted["place"] += "_shifted"
    plot_df = pd.concat(
        [discharge_down, discharge_up, discharge_up_shifted], ignore_index=True
    )
    fg = sns.FacetGrid(
        plot_df,
        col="event",
        col_wrap=2,
        hue="place",
        palette={
            down: "#467AC7",
            up + "_original": "#89AC87",
            up + "_shifted": "#298C0A",
        },
        height=fig_height,
        aspect=2,
        sharex=False,
        sharey=False,
    )
    fg.map_dataframe(sns.lineplot, x="datetime", y="discharge")
    fg.add_legend()
    fg.figure.text(1.05, 0.8, str(up_shift))
    fg.savefig(out_folder / f"diff_conc2_{up}-{down}_shift_lines.svg")


def ff_line(ff_df: pd.DataFrame, out_folder: pathlib.Path):
    """Draw accumulated mass vs volume plots as line plots. All lines are colored by event and overlap in one single ax."""
    for p, ff_df_p in tqdm(ff_df.groupby("place")):
        plot_df = (
            ff_df_p.copy()
            .reset_index()
            .rename(
                columns={"volume_ratio": "volume ratio", "load_ratio": "load ratio"}
            )
        )
        fg = sns.FacetGrid(
            plot_df.reset_index(),
            hue="event",
            palette="rocket",
            sharex=False,
            sharey=False,
            height=fig_height,
            aspect=1,
        ).map(sns.lineplot, "volume ratio", "load ratio", alpha=0.7)
        for ax in fg.axes_dict.values():
            ax.plot([0, 1], [0, 1], "--", color="gray", linewidth=1.0)
        #plt.suptitle(f"{p}: accumulated load vs volume", y=1.04)
        fg.savefig(out_folder / f"{p.replace(' ','')}_ff_plot.svg")


def ff_box_leutnant2016(ff_df: pd.DataFrame, out_folder: pathlib.Path):
    """box plot graph according to Leutnant 2016. Like ff_line but with boxplots instead of lines."""
    for p, ff_df_p in tqdm(ff_df.groupby("place")):
        volume_ratios = np.linspace(0.1, 1, 10)
        volume_ratio_idx = pd.concat(
            {
                r: (ff_df_p["volume_ratio"] - r)
                .dropna()
                .abs()
                .groupby(level=["place", "event"])
                .idxmin()
                for r in volume_ratios
            },
            names=["volume_ratio"],
        )
        plot_df = (
            pd.DataFrame(
                index=volume_ratio_idx.index,
                data={
                    "load_ratio": ff_df_p.loc[volume_ratio_idx, "load_ratio"].to_numpy()
                },
            )
            .reset_index()
            .rename(
                columns={"volume_ratio": "volume ratio", "load_ratio": "load ratio"}
            )
        )
        plt.figure(figsize=(fig_height * 1.2, fig_height))
        ax = sns.boxplot(
            plot_df,
            x="volume ratio",
            y="load ratio",
        )
        ax.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter("%.1f"))
        plt.suptitle(p, y=10.4)
        ax.get_figure().savefig(out_folder / f"{p.replace(' ','')}ff_leutnant2016.svg")


def ff_bachmann_2020(
    ff_df: pd.DataFrame, sealed_area_m2: pd.Series, out_folder: pathlib.Path
):
    """box plot graph according to Bachmann 2020. Here the volume is absolute."""
    max_mff = ff_df["mff"].quantile(0.95)
    pad = 0.05
    ylim = (-pad * max_mff, (1 + pad) * max_mff)
    for p, ff_df_p in tqdm(ff_df.groupby("place")):
        # MFF_Vol graph according to bachmann 2020
        vol_sections = sorted(
            da.volume_sections_from_raindepth(
                ff_df_p["volume_cum"], sealed_area_m2[p]
            ).unique()
        )
        vol_section_idx = {}
        for t in vol_sections:
            idx = (ff_df_p["volume_cum"] - t).abs().groupby(level=["event"]).idxmin()
            valid = ff_df_p["timestep_volume"].groupby(level=["event"]).sum() > t
            vol_section_idx[t] = idx[valid]
        vol_section_idx = pd.concat(vol_section_idx, names=["volume section"])
        plot_df = pd.DataFrame(index=vol_section_idx.index).reset_index()
        plot_df["mass first flush ratio"] = ff_df_p.loc[vol_section_idx, "mff"].values
        fg_box = sns.FacetGrid(
            plot_df, height=fig_height, aspect=1.6, sharex=True, sharey=False, ylim=ylim
        )
        fg_box.map_dataframe(
            sns.boxplot,
            x="volume section",
            y="mass first flush ratio",
            **box_kwargs,
            native_scale=True,
        )
        for ax in fg_box.axes_dict.values():
            ax.tick_params(axis="x", rotation=30)
            ax.grid(True, axis="y")
        section_size = da.calc_volume_from_raindepth(da.slice_mm, sealed_area_m2[p])
        plt.suptitle(
            f"mff at volume sections({section_size} m³), storage = {p}", y=1.04
        )
        fg_box.savefig(out_folder / f"{p.replace(' ','')}ff_bachmann2020_box.svg")
        fg_line = sns.FacetGrid(
            plot_df, height=fig_height, aspect=1.6, sharex=True, sharey=False, ylim=ylim
        )
        fg_line.map_dataframe(
            sns.lineplot,
            x="volume section",
            y="mass first flush ratio",
            estimator="mean",
            errorbar=("pi", 50),
            marker=".",
            markersize=6,
            err_style="band",
        )
        for ax in fg_line.axes_dict.values():
            ax.tick_params(axis="x", rotation=30)
            ax.hlines(
                y=1,
                xmin=0,
                xmax=vol_sections[-1],
                color="black",
                linestyles="dashed",
                linewidth=0.5,
            )
            ax.grid(True, axis="y")
        plt.suptitle(
            f"mff at volume sections ({section_size} m³), storage = {p}", y=1.04
        )
        fg_line.savefig(out_folder / f"{p.replace(' ','')}ff_bachmann2020_line.svg")


def ff_bach_2010(
    ff_df: pd.DataFrame, sealed_area_m2: pd.Series, out_folder: pathlib.Path, slice_size:float = 1
):
    """
    see:
    Peter M. Bach, David T. McCarthy, Ana Deletic,
    Redefining the stormwater first flush phenomenon,
    Water Research,
    Volume 44, Issue 8,
    2010,
    Pages 2487-2498,
    ISSN 0043-1354,
    https://doi.org/10.1016/j.watres.2010.01.022.
    (https://www.sciencedirect.com/science/article/pii/S0043135410000564)
    """
    # In the process of creatign the plots some data is collected, this is returned in the dataframe
    ff_info = pd.DataFrame()
    for storage, ff_part in ff_df.groupby("place"):
        df = ff_part.copy()
        # Calculate the rainfall depth
        df["rainfall_depth"] = da.calc_raindepth(df["volume_cum"], sealed_area_m2[storage])
        # note that this is the number if the slice, starting from 0, not the upper bound of the slice
        df["slice_num"] = np.floor(df["rainfall_depth"] / slice_size).astype(int)
        # mean concentration for a slice should be a flow-weighted sum: sum(afs*discharge*dt)/sum(sum(discharge*dt))
        # since timestep is constant, we can leave out dt. do not use transport here, otherwise units would need to be converted.
        df["afs_x_discharge"] = df["afs"]*df["discharge"]
        sg = df.groupby(["event", "slice_num"])
        slice_stats = pd.DataFrame(
            {
                "slice_size_max": sg["rainfall_depth"].max(),
                "mean_conc": sg["afs_x_discharge"].sum()/sg["discharge"].sum(),
            }
        ).reset_index()
        # filter slices with fewer that 5 values
        slice_stats["ev_count"]=slice_stats.groupby("slice_num")["event"].transform("count")
        slice_stats = slice_stats.query("ev_count >= 5").copy()
        # merge slices that are similar
        slice_stats["original_slice_num"]=slice_stats["slice_num"]
        max_slice_num = slice_stats["slice_num"].max()
        # go over all slices and assign merged_slice_num to the next slice num if the slices are similar
        # note that the slices are filtered with merged slice_num instead of original_slice_num,
        # so that next_means is compared against the whole merge groupp.
        for i in range(max_slice_num):
            i_mask = slice_stats["slice_num"]==i
            i_means = slice_stats.loc[i_mask,"mean_conc"].to_numpy()
            next_means = slice_stats.loc[slice_stats["slice_num"]==(i+1),"mean_conc"].to_numpy()
            # Perform the Wilcoxon Rank Sum Test
            _, p_value = scipy.stats.ranksums(
                i_means, next_means
            )
            # Check if the p-value is greater than 0.05
            if p_value > 0.05:
                slice_stats.loc[i_mask,"slice_num"]=i+1
        # slice_stats is mostly accessed by slice num, set that as index
        slice_stats = slice_stats.set_index("slice_num")
        # after the merge slice num is not a simple range, use slice_nums to iterate over all values
        slice_nums=np.array(sorted(slice_stats.index.unique()))
        # plot
        fig, ax = plt.subplots()
        # Upper bounds of the slices, used for x-ticks
        slice_ubounds = slice_nums*slice_size+slice_size
        # Widths of the slices and corresponding box plots.
        # Calculate the widths based on the differences to previous values, just the value for the first entry
        widths = slice_ubounds - np.roll(slice_ubounds,1)
        widths[0] = slice_ubounds[0]
        # x-positions of the plots, each boxplot is horizontally centered around this position.
        # These positions are also used for extra notes.
        positions = slice_ubounds - widths/2
        # the values used in the boxplot, list with one array for each boxplot
        boxplot_data = []
        for i in slice_nums:
            boxplot_data.append(slice_stats.loc[[i], "mean_conc"].to_numpy())
        # subtract some padding from the boxplot widths, so that they have a gap.
        wpad = 0.05
        ax.boxplot(
            boxplot_data,
            positions=positions,
            widths=widths - wpad * 2,
            boxprops=dict(facecolor="#65b22e"),
            flierprops=dict(
                marker="o",
                markersize=1,
                markerfacecolor="none",
                markeredgecolor="black",
            ),
            medianprops=dict(color="black"),
            **box_kwargs,
            patch_artist=True,
        )
        ax.set_xticks(slice_ubounds)
        ax.set_xticklabels([str(s) for s in slice_ubounds])
        # Set axis limits
        ax.set_xlim(-0.5, max(slice_ubounds) + 0.5)
        ax.set_ylim(-10, 1000)
        
        # Add number of  values in a slice above each boxplot
        for i in range(len(slice_nums)):
            ax.text(
                positions[i],
                1.007,
                f"n={len(boxplot_data[i])}",
                transform=ax.get_xaxis_transform(),
                horizontalalignment="center",
                size="small",
                weight="semibold",
                color="black",
                rotation=45,
            )
        
        # Set axis labels
        ax.set_ylabel("TSS Concentration (mg/l)")
        ax.set_xlabel("Cumulative runoff depth (mm)")
        
        # Reduces the borders to fit in the labels
        fig.tight_layout()
        #fig.suptitle(storage, y=1.06)
        # Save the plot
        file_path = (
            out_folder
            / f"{storage.replace(' ','')}_bachmann_runoffdepth_conc_1mm_combinedslices_old.svg"
        )
        fig.savefig(file_path, format="svg", bbox_inches="tight")

        # plot load
        df["rainfall_depth_ceil"] = (
            np.ceil(df["rainfall_depth"] / slice_size) * slice_size
        )
        max_x = max(slice_ubounds) + 0.5
        df_cut = df.query("rainfall_depth<=@max_x")

        # using the mean of the accumulated load over each event is confusing, because this mean is not monotonic increasing
        # instead we sum up the changes in load for each rainfall_depth_ceil section per event
        # and than take the mean of the change over the events.
        # finally the changes are accumulated with cumsum().
        load_mean = (
            df_cut.groupby(["event", "rainfall_depth_ceil"])["timestep_load"]
            .sum()
            .groupby("rainfall_depth_ceil")
            .mean()
            .cumsum()
            .to_numpy()
        )
        load_depth_slices = df_cut["rainfall_depth_ceil"].unique()
        # add 0,0
        load_depth_slices = np.concatenate([[0], load_depth_slices])
        load_mean = np.concatenate([[0], load_mean])
        ax_load = ax.twinx()
        ax_load.plot(load_depth_slices, load_mean, color="black", alpha=0.6)
        ax_load.set_ylim(-0.01 * load_mean.max(), 1.01 * load_mean.max())
        ax_load.set_ylabel("cumulated mean TSS load (kg)")
        file_path_load = (
            out_folder
            / f"{storage.replace(' ','')}_bachmann_runoffdepth_conc_1mm_combinedslices_load.svg"
        )
        fig.savefig(file_path_load, format="svg", bbox_inches="tight")

        # Extract data for the first and last slices
        first_conc = slice_stats.loc[[slice_nums[0]], "mean_conc"]
        last_conc = slice_stats.loc[[slice_nums[-1]], "mean_conc"]
        if len(slice_nums) >= 3:
            before_last_data = slice_stats.loc[[slice_nums[-2]]]
        else:
            before_last_data = pd.DataFrame(index=[0], columns=slice_stats.columns, data=np.nan)
        # Perform the Wilcoxon Rank Sum Test
        _, p_value = scipy.stats.ranksums(first_conc.to_numpy(), last_conc.to_numpy())
        ff_info.loc[storage, "bach_init_conc"] = first_conc.median()
        ff_info.loc[storage, "bach_init_conc_mean"] = first_conc.mean()
        ff_info.loc[storage, "bach_bg_conc"] = last_conc.median()
        ff_info.loc[storage, "bach_ff_strength"] = p_value
        ff_info.loc[storage, "bach_before_bg_conc"] = before_last_data["mean_conc"].median()
        ff_info.loc[storage, "bach_max_depth"] = slice_stats["slice_size_max"].max()
        ff_info.loc[storage, "bach_before_last_depth"] = before_last_data[
            "slice_size_max"
        ].max()
    return ff_info
